
#include <XSUtil/Numerics/RandomGenerator.h>

#include <XSUtil/Numerics/Gamma.h>

#include <cmath>
#include <algorithm>

#include <iostream>
#include <fstream>


namespace Numerics {

void PoissonRand(RealArray& poissonMeans)
{
   size_t N (poissonMeans.size());  
   static const Real PI = 4.*std::atan(1.);  
   const DefaultRandomGenerator& randGen = DefaultRandomGenerator::instance();
        
// Function to return NM random Poisson numbers whose means are
// given by XM(NM). Algorithm from Press etal and random number
// generator from CERN.

   float oldm(-1); 
   float g(0);
   float sq(0);
   float alxm(0);
   float y(0);

   for (size_t i = 0; i < N; ++i)
   {
      float xm (poissonMeans[i]);
      float em(0);
      float random(0);
      if ( xm < 12. ) 
      {
         if (xm != oldm)
         {
            oldm = xm;
            g = exp(-xm);       
         } 
         em = 0;
         randGen.getRandom(&random,1);
         float t (random);
         while ( t > g )
         {
            ++em;
            randGen.getRandom(&random,1);
            t *= random;
         }
      }
      else
      {
         GammaLN G;
         if ( xm != oldm )
         {
            oldm = xm;
            sq = sqrt(2.*xm);
            alxm = log(xm);
            g = xm * alxm - G(xm + 1);

         }
         bool done (false);
         while ( !done )
         {
            em = -1;
            while ( em < 0 )
            {
               randGen.getRandom(&random,1);
               y = tan( PI*random );
               em = sq * y + xm;
            } 
            em = int(em);
            float t = 0.9*(1 + y*y)*exp(em*alxm - G(em + 1) - g);
            randGen.getRandom(&random,1);
            done = random <= t;      
         }
      }

      poissonMeans[i] = em;
      oldm = xm;
   }

}                
                
}
